import Flutter
import UIKit
import AliyunOSSiOS
import MobileCoreServices

public class SwiftAliyunOSSPlugin: NSObject, FlutterPlugin, UIDocumentPickerDelegate, UINavigationControllerDelegate {
  
  let channel: FlutterMethodChannel
  var flutterViewController: UIViewController?
  var client: OSSClient!
  var bucketName: String!
  
  init(channel: FlutterMethodChannel) {
    self.channel = channel
  }
  
  public static func register(with registrar: FlutterPluginRegistrar) {
    let channel = FlutterMethodChannel(name: "com.wombat/aliyun_oss_plugin", binaryMessenger: registrar.messenger())
    let instance = SwiftAliyunOSSPlugin(channel: channel)
    registrar.addMethodCallDelegate(instance, channel: channel)
    
    let eventChannel = FlutterEventChannel(name: "com.wombat/aliyun_oss_stream_plugin", binaryMessenger: registrar.messenger())
    let eventInstance = SwiftEventChannelHandler()
    eventChannel.setStreamHandler(eventInstance)
  }
  
  public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
    switch call.method{
    case "init":
      initClient(call)
    case "putObject":
      self.bucketName = (call.arguments as! [String: String])["bucketName"]
      print("上传文件: \(bucketName)")
      selectFile()
    default:
      result(FlutterMethodNotImplemented)
    }
  }
  
  func initClient(_ call: FlutterMethodCall) {
    let params = call.arguments as! [String: Any?]
    let endpoint = params["endpoint"] as! String
    let accessKeyId = params["accessKeyId"] as! String
    let secretKeyId = params["secretKeyId"] as! String
    let securityToken = params["securityToken"] as? String
    
    var credential: OSSCredentialProvider
    if (securityToken == nil) {
      credential = OSSPlainTextAKSKPairCredentialProvider(plainTextAccessKey: accessKeyId, secretKey: secretKeyId)
    } else {
      credential = OSSStsTokenCredentialProvider(accessKeyId: accessKeyId, secretKeyId: secretKeyId, securityToken: securityToken!)
    }
    
    client = OSSClient(endpoint: endpoint, credentialProvider: credential)
    
  }
  
  func selectFile() {
    var documentPicker: UIDocumentPickerViewController = UIDocumentPickerViewController(documentTypes: ["public.item"], in: UIDocumentPickerMode.import)
    documentPicker.delegate = self
    documentPicker.modalPresentationStyle = UIModalPresentationStyle.fullScreen
    if (flutterViewController == nil) {
      flutterViewController = UIApplication.shared.keyWindow?.rootViewController;
    }
    flutterViewController?.present(documentPicker, animated: true, completion: nil);
    
  }
  
  public func documentPicker(_ controller: UIDocumentPickerViewController, didPickDocumentsAt urls: [URL]) {
    guard let fileURL = urls.first else {
      return
    }
    DispatchQueue(label: "calcuate base64Md5", qos: .background).async {
      print("正在计算md5中...")
      let start = CACurrentMediaTime()
      guard let md5: String = OSSUtil.base64Md5(forFileURL: fileURL) else {
        return
      }
      let elapsed = CACurrentMediaTime() - start
      print("计算md5: \(md5), 线程: \(Thread.current), 耗时: \(elapsed)")
      DispatchQueue.main.async {
        self.multipartUpload(fileURL: fileURL, md5: md5)
        //        self.putObject(fileURL: fileURL, md5: md5)
      }
    }
  }
  
  func fileExt(_ fileURL: URL) -> String{
    var ext = fileURL.absoluteString.lowercased()
    
    if #available(iOS 11.0, *) {
      if (ext.contains(".")){
        ext = String(ext[ext.lastIndex(of: ".")!...])
      }
    }
    return ext
  }
  
  public func documentPickerWasCancelled(_ controller: UIDocumentPickerViewController) {
    flutterViewController?.dismiss(animated: true, completion: nil)
  }
  
  func showResult(task: OSSTask<AnyObject>?, md5: String? = nil, start: CFTimeInterval? = nil) -> Void {
    
    if (start != nil) {
      let elapsed = CACurrentMediaTime() - start!
      print("上传结束: \(md5), 耗时: \(elapsed)")
    }
    if (task?.error != nil) {
      let error: NSError = (task?.error)! as NSError
      print("上传出错: \(error)")
    } else {
      let result = task?.result
      print("上传成功: \(result)")
      progressCallback(bytesSent: -1, totalBytesSent: -1, totalBytesExpectedToSend: -1, finish: true)
    }
  }
  
  func progressCallback(bytesSent: Int64, totalBytesSent: Int64, totalBytesExpectedToSend: Int64,finish: Bool) {
    channel.invokeMethod("upload_progress", arguments: [
      "bytesSent": bytesSent,
      "totalBytesSent": totalBytesSent,
      "totalBytesExpectedToSend": totalBytesExpectedToSend,
      "finish": finish
    ]);
  }
  
  func putObject(fileURL: URL, md5: String) -> Void {
    print("上传: \(Thread.current)")
    let ext = fileExt(fileURL)
    let request = OSSPutObjectRequest()
    request.uploadingFileURL = fileURL
    request.bucketName = bucketName
    request.objectKey = "\(md5)\(ext)"
    print("文件名称objectKey: \(md5)\(ext)")
    request.uploadProgress = { (bytesSent: Int64, totalBytesSent: Int64, totalBytesExpectedToSend: Int64) -> Void in
      print("进度 bytesSent:\(bytesSent),totalBytesSent:\(totalBytesSent),totalBytesExpectedToSend:\(totalBytesExpectedToSend)");
      self.progressCallback(bytesSent: bytesSent, totalBytesSent: totalBytesSent, totalBytesExpectedToSend: totalBytesExpectedToSend, finish: false)
    };
    
    let task = client.putObject(request)
    task.continue({ (t) -> Any? in
      self.showResult(task: t)
    }).waitUntilFinished()
  }
  
  func multipartUpload(fileURL: URL, md5: String) -> Void {
    let start = CACurrentMediaTime()
    print("开始上传: \(md5)")
    let ext = fileExt(fileURL)
    let request = OSSMultipartUploadRequest()
    request.uploadingFileURL = fileURL
    request.bucketName = bucketName
    request.objectKey = "\(md5)\(ext)"
    request.partSize = 102400;
    request.uploadProgress = { (bytesSent: Int64, totalBytesSent: Int64, totalBytesExpectedToSend: Int64) -> Void in
      print("bytesSent:\(bytesSent),totalBytesSent:\(totalBytesSent),totalBytesExpectedToSend:\(totalBytesExpectedToSend)");
      self.progressCallback(bytesSent: bytesSent, totalBytesSent: totalBytesSent, totalBytesExpectedToSend: totalBytesExpectedToSend, finish: false)
    };
    
    let task = client.multipartUpload(request)
    task.continue({ (t) -> Any? in
      self.showResult(task: t, md5: md5, start: start)
    }).waitUntilFinished()
  }
  
  func resumableUpload() -> Void {
    var request = OSSResumableUploadRequest()
    request.uploadingFileURL = Bundle.main.url(forResource: "wangwang", withExtension: "zip")!
    request.bucketName = bucketName
    request.deleteUploadIdOnCancelling = false;
    request.objectKey = "wangwang(swift).zip"
    let cacheDir =  NSSearchPathForDirectoriesInDomains(FileManager.SearchPathDirectory.cachesDirectory, FileManager.SearchPathDomainMask.userDomainMask, true).first
    request.recordDirectoryPath = cacheDir!
    request.partSize = 102400;
    request.uploadProgress = { (bytesSent: Int64, totalBytesSent: Int64, totalBytesExpectedToSend: Int64) -> Void in
      print("bytesSent:\(bytesSent),totalBytesSent:\(totalBytesSent),totalBytesExpectedToSend:\(totalBytesExpectedToSend)");
      if totalBytesSent > (totalBytesExpectedToSend / 2) {
        request.cancel()
      }
    }
    
    var task = client.resumableUpload(request)
    task.continue({ (t) -> Any? in
      print("Error: \(String(describing: t.error))")
      return nil
    }).waitUntilFinished()
    
    request = OSSResumableUploadRequest()
    request.uploadingFileURL = Bundle.main.url(forResource: "wangwang", withExtension: "zip")!
    request.bucketName = bucketName
    request.objectKey = "wangwang(swift).zip"
    request.partSize = 102400;
    request.deleteUploadIdOnCancelling = false;
    request.recordDirectoryPath = cacheDir!
    request.uploadProgress = { (bytesSent: Int64, totalBytesSent: Int64, totalBytesExpectedToSend: Int64) -> Void in
      print("bytesSent:\(bytesSent),totalBytesSent:\(totalBytesSent),totalBytesExpectedToSend:\(totalBytesExpectedToSend)");
    }
    
    task = client.resumableUpload(request)
    task.continue({ (t) -> Any? in
      self.showResult(task: t)
      return nil
    }).waitUntilFinished()
  }
  
  func sequentialMultipartUpload() {
    let request = OSSResumableUploadRequest()
    request.bucketName = bucketName;
    request.objectKey = "sequential-swift-multipart";
    request.uploadingFileURL = Bundle.main.url(forResource: "wangwang", withExtension: "zip")!
    request.deleteUploadIdOnCancelling = false
    request.crcFlag = OSSRequestCRCFlag.open
    let filePath = Bundle.main.path(forResource: "wangwang", ofType: "zip")
    request.contentSHA1 = OSSUtil.sha1(withFilePath: filePath)
    
    let task = client.sequentialMultipartUpload(request)
    task.continue({ (t) -> Any? in
      self.showResult(task: t)
      
      return nil
    }).waitUntilFinished()
  }
  
}
